/* MODULE CMDCRN */

#define MODULE
#include <linux/module.h>
#include <linux/kernel.h>

#include <rtl_sched.h>

#define FIVE_SECONDS 5000000000LL
static	enum {stoppedInitial, running, stoppedFinal} Chronostatus;
static RT_TASK Chrono;

void CommandChrono_Put(char command)
{
	static RT_TASK *ackn = 0;
	unsigned int put = 'c';
	unsigned int msg;

	if (ackn != &Chrono) {
		ackn = rt_rpc(&Chrono, put, &msg);
	}
	if ((Chronostatus == running) != (command == 'C')) {
		rt_send(&Chrono, (unsigned int)command);
	}
}

void CommandChrono_Get(char *command)
{
	static RT_TASK *ackn = 0;
	unsigned int get = 'd';
	unsigned int msg;

	if (ackn != &Chrono) {
		ackn = rt_rpc(&Chrono, get, &msg);
	}
	rt_receive(&Chrono, &msg);
	*command = (char)msg;
}

void CommandChrono_task(int t)
{
	RTIME fiveSeconds = nano2count(FIVE_SECONDS);
	unsigned int command;
	unsigned int buffered = 0;
	unsigned int C = 'C';
	unsigned int R = 'R';
	int ackn = 0;
	RT_TASK *get = (RT_TASK *)0, *put = (RT_TASK *)0, *task;

	Chronostatus = stoppedInitial;
	while (ackn != ('c' + 'd')) {
		task = rt_receive((RT_TASK *)0, &command);
		switch (command) {
			case 'd':
				get = task;
				ackn += command;
				break;
			case 'c':
				put = task;
				ackn += command;
				break;
		}
	} 
	rt_return(put, command);
	rt_return(get, command);

	while(1) {
		switch (Chronostatus) {
			case stoppedInitial:
				if (buffered) {
					command = buffered;
					buffered = 0;
				} else {
					rt_receive(put, &command);
				}	
				Chronostatus = running;
				break;
			case running:
				if (rt_receive_if(put, &command)) {
					if (command == 'E') {
						Chronostatus = stoppedFinal;
					}
				} else {
					command = C;
				}
				break;
			case stoppedFinal:
				Chronostatus = stoppedInitial;
				if (rt_receive_timed(put, &command, fiveSeconds) > 0) {
					buffered = command;
				}
				command = R;
				break;
		}
		rt_send(get, command);
	}
}

int init_module(void)
{
	rt_task_init(&Chrono, &CommandChrono_task, 0, 2000, 0, 0, 0);
	rt_task_resume(&Chrono);
	return 0;
}

void cleanup_module(void)
{
	rt_task_delete(&Chrono);
}
