/* MODULE ClockChrono */

#define MODULE
#include <linux/module.h>
#include <linux/rt_sched.h>
#include <linux/rtf.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>

#include <CommandClock.h>
#include <CommandChrono.h>
#include <MenageHmsh.h>
#include <Display.h>
#include <Type.h>

#define Keyboard 0
#define Screen   1

void ClockChrono_Read(void)
{
	char ch;

	while(1) {
		rtf_get(Keyboard, &ch, 1);
		ch = toupper(ch);
		switch(ch) {
			case 'T': case 'R': case 'H': case 'M': case 'S':
				CommandClock_Put(ch);
				break;
			case 'C': case 'I': case 'F':
				CommandChrono_Put(ch);
			default:
				break;
		}
	}
}

void ClockChrono_Clock(void)
{
	const hundredthes = FALSE;
	MenageHmsh_tHour hour;
	MenageHmsh_tChain11 hourChain;
	char command;
	BOOLEAN display;

	MenageHmsh_Initialise(hour);
	MenageHmsh_Convert(hour, hundredthes, hourChain);
	Display_PutHour(hourChain);
	while(1) {
		CommandClock_Get(command);
		switch(command) {
			case 'R':
				rt_sleep(nano2counts(10000000LL));
				MenageHmsh_PlusOneUnit(&hour, display);
				break;
			case 'T': 
				ManageHmsh_InitialiseHundredthes(&hour);
				display = FALSE;
				break;
			case 'H': 
				MenageHmsh_AdvanceHours(&hour);
				display = TRUE;
				break;
			case 'M':
				MenageHmsh_AdvanceMinutes(&hour);
				display = TRUE;
				break;
			case 'S':
				MenageHmsh_AdvanceSeconds(&hour);
				display = TRUE;
			}
		if (display) {
			MenageHmsh_Convert(hour, hundredthes, hourChain);
			Display_PutHour(hourChain);
		}
	}
}

void ClockChrono_Chrono(void)
{
	MenageHmsh_tHour times;			
	MenageHmsh_tChain11 timesChain;
	BOOLEAN Intermediatetimes;
	MenageHmsh_tHour endIntermediateTimes;
	BOOLEAN display;
	BOOLEAN hundredthes;
	char command;

	command = 'R';
	while(1) {
		switch(command) {
			case 'R':
				MenageHmsh_Initialise(&times);
				display = TRUE;
				hundredthes = FALSE;
				Intermediatetimes = FALSE;
				break;
			case 'C':
				rt_sleep(nano2counts(10000000LL));
				MenageHmsh_PlusOneUnit(&times, display);
				if (Intermediatetimes) {
					Intermediatetimes = !MenageHmshEqual(
						times, endIntermadiateTimes);
					display = !Intermediatetimes;
					hundredthes = FALSE;
				}
				break;
			case 'I':
				Intermediatetimes = TRUE;
				endIntermadiateTimes = times;
				MenageHmsh_PlusNseconds(3, 
							&endIntermadiateTimes);
				display = TRUE;
				hundredthes = TRUE;
				break;
			case 'F':
				display = TRUE;
				hundredthes = TRUE;
		}
		if (display) {
			MenageHmsh_Convert(times, hundredthes, timesChain);
			Display_PutTimes(timesChain);
		}
		CommandChrono_Get(command);
	}
}

void ClockChrono_Write(void)
{
	int i;
	MenageHmsh_tChain11 chain;

	rtf_put(Screen, '\n', 1);
	while(1) {
		Display_Get(chain, receiver);
		rtf_put(Screen, '\n', 1);
		if (receiver == Display_destChrono) {
			rtf_put(Screen, '\n', 1);
		}
		for (i = 0; i < sizeof(chain); i++) {
			rtf_put(Screen, chain[i], 1);
		}
	}
}

RT_TASK read;
RT_TASK clock;
RT_TASK chrono;
RT_TASK write;

int init_module(void)
{
	rtf_create(Keyboard, 2000);
	rtf_create(Screen, 2000);
	rt_task_init(&read, ClockChrono_Read, 0, 2000, 0, 0);
	rt_task_init(&clock, ClockChrono_Clock, 0, 2000, 0, 0);
	rt_task_init(&chrono, ClockChrono_Chrono, 0, 2000, 0, 0);
	rt_task_init(&write, ClockChrono_Write, 0, 2000, 0, 0);
	rt_start_timer((int)nano2count(100000LL));
	return 0;
}

void cleanup_module(void)
{
	rt_stop_timer(();
	rtf_destroy(Keyboard);
	rtf_destroy(Screen);
	rt_task_delete(&read);
	rt_task_delete(&clock);
	rt_task_delete(&chrono);
	rt_task_delete(&write);
}


/* MODULE CommandClock */

#include <rt_sched.h>

enum (stopped, running) Clockstatus;
char buffer;
SEM notEmpty, notFull, mutex;

void CommandClock_Put(char command)
{
	rt_sem_wait(&notFull);
	rt_sem_wait(&mutex);
	if ((Clockstatus == running) == (command == 'T')) {
		buffer = command;
		rt_sem_signal(&mutex);
		rt_sem_signal(&notEmpty);
	} else {
		rt_sem_signal(&mutex);
	}
}

void CommandClock_Get(char *command)
{
	switch (Clockstatus) {
		case stopped:
			rt_sem_wait(&notEmpty);
			rt_sem_wait(&mutex);
			*command = buffer;
			if (*command == 'R') {
				Clockstatus = running;
			}
			rt_sem_signal(&mutex);
			rt_sem_signal(&notEmpty);
			break;
		case running:
			rt_sem_wait(&mutex);
			*command = buffer;
			if (*command == 'T') {
				Clockstatus = stopped;
			}
			rt_sem_signal(&mutex);
			rt_sem_signal(&notEmpty);
			break;
	}
}

int init_module(void)
{
	rt_sem_init(&notEmpty, 0);
	rt_sem_init(&notFull, 1);
	rt_sem_init(&mutex, 1);
	Clockstatus = stopped;
	return 0
}

void cleanup_module(void)
{
}

/* MODULE CommandChrono */

#include <rt_sched.h>

const fiveSeconds = 5000000000LL;
enum (stoppedInitial, running, stoppedFinal) Chronostatus;
char buffer;
SEM notEmpty, notFull, mutex;

void CommandChrono_Put(char command)
{
	rt_sem_wait(&notFull);
	rt_sem_wait(&mutex);
	if ((Chronostatus == running) != (command == 'C')) {
		buffer = command;
		rt_sem_signal(&mutex);
		rt_sem_signal(&notEmpty);
	} else {
		rt_sem_signal(&mutex);
	}
}

void CommandChrono_Get(char *command)
{
	switch (Chronostatus) {
		case stoppedInitial:
			rt_sem_wait(&notEmpty);
			rt_sem_wait(&mutex);
			*command = buffer;
			Chronostatus = running;
			rt_sem_signal(&mutex);
			rt_sem_signal(&notFull);
			break;
		case running:
			rt_sem_wait(&mutex);
			if (rt_sem_wait_if(&notFull) > 0) {
				*command = 'C';
			} else {
				*command = buffer;
			}
			if (*command == 'F') {
				Chronostatus = stoppedFinal;
			}
			rt_sem_signal(&mutex);
			rt_sem_signal(&notFull);
			break;
		case stoppedFinal:
			rt_sem_wait_timed(&notEmpty, nano2counts(fiveSeconds));
			rt_sem_wait(&mutex);
			*command = 'R';
			Chronostatus = stoppedInitial;
			rt_sem_signal(&mutex);
			break;
	}
}

int init_module(void)
{
	rt_sem_init(&notEmpty, 0);
	rt_sem_init(&notFull, 1);
	rt_sem_init(&mutex, 1);
	Clockstatus = stoppedInitial;
	return 0;
}

void cleanup_module(void)
{
}

/* MODULE MenageHmsh */

#include <Type.h>

struct {int hours, minutes, secondes, hundredthes;} tHour;
char tChain11[11];
char digits[10] = {'0','1','2','3','4','5'.'6','7','8','9'};


void MenageHmsh_Initialise(tHour *h)
{
	h->hours = h->minutes = h->seconds = h->hundredthes = 0;
}

void MenageHmsh_Initialise_Hundredthes(tHour *h)
{
	h->hundredthes = 0;
}

BOOLEAN MenageHmsh_Equal(tHour h1, h2)
{
	return 	   (h1.hours == h2.hours)
		&& (h1.minutes == h2.minutes)
		&& (h1.seconds == h2.seconds)
		&& (h1.hundredthes == h2.hundredthes);
}

void MenageHmsh_AdvanceHours(tHour *h)
{
	h->hours = (h->hours + 1)%24;
}

void MenageHmsh_AdvanceMinutes(tHour *h)
{
	h->minutes = (h->minutes + 1)%60;
}

void MenageHmsh_AdvanceSeconds(tHour *h)
{
	h->seconds = (h->seconds + 1)%60;
}

void MenageHmsh_PlusOneSecond(tHour *h)
{
	MenageHmsh_AdvanceSeconds(h);
	if (h->seconds == 0) {
		MenageHmsh_AdvanceMinutes(h);
		if (h->minutes == 0) {
			MenageHmsh_AdvanceHours(h);
		}
	}
}

void MenageHmsh_PlusNSeconds(int n, tHour *h)
{
	int i;
	for (i = 1; i <= n; i++) {
		MenageHmsh_PlusOneSecond(h);
	}
}

void MenageHmsh_PlusOneUnit(tHour *h, BOOLEAN *reportSeconds)
{
	h->hundredthes = (h->hundredthes + 1)%100; 
	*reportSeconds = h->hundredthes == 0; 
	if (*reportSeconds) {
		MenageHmsh_PlusOneSecond(h);
	}
}

void MenageHmsh_Convert(tHour h, BOOLEAN hundredthes, tChaine11 chain)
{
	chain[0] = digits[h.hours/10];
	chain[1] = digits[h.hours%10];
	chain[2] = ':';
	chain[3] = digits[h.minutes/10];
	chain[4] = digits[h.minutes%10];
	chain[5] = ':';
	chain[6] = digits[h.seconds/10];
	chain[7] = digits[h.seconds%10];
	if (hundredthes) {
		chain[08] = '.';
		chain[09] = digits[h.hundredthes/10];
		chain[10] = digits[h.hundredthes%10];
	} else {
		chain[08] = chain[09] =	chain[10] = ' ';
	}
}

/* MODULE Display */

#include <rt_sched.h>
#include <MenageHmsh.h>

enum (empty, full) HourBufferstatus, TimesBufferstatus;
tChain11 hour;
tChain11 times;
SEM notEmpty, mutex;

void Display_PutHour(tChain11 chain)
{
	rt_sem_wait(&mutex);
	hour = chain;
	HourBufferstatus = full;
	rt_sem_signal(&mutex);
	rt_sem_signal(&notEmpty);
}

void Display_PutTimes(tChain11 chain)
{
	rt_sem_wait(&mutex);
	times = chain;
	TimesBufferstatus = full;
	rt_sem_signal(&mutex);
	rt_sem_signal(&notEmpty);
}

void Display_Get(tChain11 chain, tDest *receiver)
{
	rt_sem_wait(&notEmpty);
	rt_sem_wait(&mutex);
	if (TimesBufferstatus == full) {
		receiver = destChrono; 
		chain = times;
		TimesBufferstatus = empty;
	} else {
		receiver = destClock; 
		chain = hours;
		HourBufferstatus = empty;
	}
	rt_sem_signal(&mutex);
}

int init_module(void)
{
	rt_sem_init(&notEmpty, 0);
	rt_sem_init(&mutex, 1);
	TimesBufferstatus = empty;
	HourBufferstatus = empty;
	return 0;
}

void cleanup_module(void)
{
}

/* DEFINITION CommandClock */

extern void CommandClock_Put(char command);
extern void CommandClock_Get(char *command);

/* DEFINITION CommandChrono */

extern void CommandChrono_Put(char command);
extern void CommandChrono_Get(char *command);

/* DEFINITION MenageHmsh */

typedef char ch[11] tChain11;
typedef struct {int hours, minutes, seconds, hundredthes;} tHour;

extern void  MenageHmsh_Initialise(tHour *h);
extern void  MenageHmsh_InitialiseHundredthes(tHour *h);
extern void  MenageHmsh_AdvanceHours(tHour *h);
extern void  MenageHmsh_AdvanceMinutes(tHour *h);
extern void  MenageHmsh_AdvanceSeconds(tHour *h);
extern void  MenageHmsh_PlusNSeconds(int n, tHour *h);
extern void  MenageHmsh_PlusOneUnit(tHour *h, BOOLEAN *CarrySeconds);
extern void  MenageHmsh_Convert(tHour h, BOOLEAN *hundredthes, tChain11 chain);
extern BOOLEAN MenageHmsh_Equal(tHour h1, h2);

/* DEFINITION Display */

#include <MenageHmsh.h>

typedef enum (destClock, destChrono) tDest;

extern void Display_PutHour(tChain11 chain);
extern void Display_PutTimes(tChain11 chain);
extern void Display_Get(tChain11 *chain, tDest *receiver);

/* DEFINITION Type */

#define BOOLEAN int
#define TRUE 1
#define FALSE 0


/* KEYBOARD */

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/rtf.h>
#include <asm/rt_time.h>
#include "control.h"

int main()
{
	int Keyboard, Screen;
	char ch;
	
	if ((Keyboard = open("/dev/rtf0", O_WRONLY)) < 0) {
		fprintf(stderr, "Error opening /dev/rtf0\n");
		exit(1);
	}

	if ((Screen = open("/dev/rtf1", O_RDONLY)) < 0) {
		fprintf(stderr, "Error opening /dev/rtf1\n");
		exit(1);
	}

	while((ch = getch()) != 'e') {
		write(Keyboard, &ch, 1);
	}

	while(1) {
		read(Screen, &ch, 1);
		printf("%c", ch);
	}
}