#define MODULE
#include <linux/module.h>
#include <asm/rt_irq.h>
#include <linux/mc146818rtc.h>

#include <rtl_fifo.h>
#include <rtl_sched.h>

#define ONESHOT

#define TICK_PERIOD 125000

#define STACK_SIZE 4000

RT_TASK task;

static int filter(int x)
{
	static int oldx;
	int ret;

	if (x & 0x80) {
		x = 382 - x;
	}
	ret = x > oldx;
	oldx = x;
	return ret;
	
}


void intr_handler(int t) {
	char data;
	char temp;
	while(1) {
		if (rtf_get(0, &data, 1) > 0) {
			data = filter(data);
			temp = inb(0x61);            
			temp &= 0xfd;
			temp |= (data & 1) << 1;
			outb(temp,0x61);
		}
		rt_task_wait_period();
	}
}


int init_module(void)
{
	RTIME now, tick_period;
	rtf_create(0, 4000);
	rt_task_init(&task, intr_handler, 0, STACK_SIZE, 0, 0, 0);
#ifdef ONESHOT
	rt_set_oneshot_mode();
#endif
	tick_period = start_rt_timer(nano2count(TICK_PERIOD));
	now = rt_get_time() + tick_period;
	rt_task_make_periodic(&task, now, tick_period);
	return 0;
}


void cleanup_module(void)
{
	stop_rt_timer();
	rtf_destroy(0);
	rt_task_delete(&task);
}
